﻿$PSCommandPathParent = "$(Split-Path -Parent -Path ([Environment]::GetCommandLineArgs()[0]))\" # for executable compilation

write-host "LOADING MODULES" -ForegroundColor Cyan

# Load NavAdminTool
#===========================================================================
try {    
    Import-Module "$($PSCommandPathParent)Cloud.Ready.Software.NAV.light" | out-null
    Import-NavModules | out-null
} catch {
    Write-Host "Seem to have problem to load Nav.Managment module. Check your system requierement."
}

# Load external SQLPS for old powershell< version
#===========================================================================
try {
    Import-Module SQLPS -DisableNameChecking | out-null
} catch {
    try {
        invoke-expression -Command "$($PSCommandPathParent)SQLPS\SQLPS.ps1"
        invoke-expression -Command "$($PSCommandPathParent)SQLPS\SqlPsPostScript.ps1"
    } catch {
        Write-Host "Seem to have problem to load SQLPS module. Check your system requirement."
    }
}

# Hide powershell console functionnality
#===========================================================================
Add-Type -Name Window -Namespace Console -MemberDefinition '
[DllImport("Kernel32.dll")]
public static extern IntPtr GetConsoleWindow();
[DllImport("user32.dll")]
public static extern bool ShowWindow(IntPtr hWnd, Int32 nCmdShow);'
function Show-Console
{
    $consolePtr = [Console.Window]::GetConsoleWindow()
    [Console.Window]::ShowWindow($consolePtr, 4)
}
function Hide-Console
{
    $consolePtr = [Console.Window]::GetConsoleWindow()
    [Console.Window]::ShowWindow($consolePtr, 0)
}
Hide-Console | out-null
$global:ConsoleShow = $false

# Load and format xaml
#===========================================================================
[void][System.Reflection.Assembly]::LoadWithPartialName('presentationframework')
[xml]$XAML = Get-Content "$($PSCommandPathParent)UI_CopyDatabase.xaml"
#Read XAML
$reader=(New-Object System.Xml.XmlNodeReader $xaml)
try{$global:Form=[Windows.Markup.XamlReader]::Load( $reader )}
catch{ Write-Host "Unable to load Windows.Markup.XamlReader. Double-check syntax and ensure .net is installed." -ForegroundColor Red }
$xaml.SelectNodes("//*[@Name]") | %{Set-Variable -Name "UI_$($_.Name)" -Value $global:Form.FindName($_.Name)}
write-host "LOAD FORM UI_CopyDatabase.xaml" -ForegroundColor Cyan
get-variable UI_*


# Load "message box" form
#=========================================================================== 
function ShowMessageBox([string]$text,[int]$type) {
    # Load the msgbox form :
    [xml]$XAML = Get-Content "$($PSCommandPathParent)UI_MessageBox.xaml"
    #Read XAML
    $reader=(New-Object System.Xml.XmlNodeReader $xaml)
    try{$global:MsgBox=[Windows.Markup.XamlReader]::Load( $reader )}
    catch{ Write-Host "Unable to load Windows.Markup.XamlReader. Double-check syntax and ensure .net is installed." -ForegroundColor Red }
    $xaml.SelectNodes("//*[@Name]") | %{Set-Variable -Name "MsgBox_$($_.Name)" -Value $global:MsgBox.FindName($_.Name)}

    # Load logo on type (0 = info 1 = error)
    if ($type -eq 0) {
        $MsgBox_image.Source = "$($PSCommandPathParent)info.png"
    }
    if ($type -eq 1) {
        $MsgBox_image.Source = "$($PSCommandPathParent)error.png"
    }

    # Show text
    $MsgBox_textBlock_Msg.Text = $text

    # Close by ok button 
    $MsgBox_button_OK.Add_Click({
        $global:MsgBox.Close()
    })

    Write-Host $text
    $global:MsgBox.ShowDialog() | out-null
}

# Refresh process status
function RefreshStatus([string]$texte) {
    $UI_textBlock_info.Text = $texte
    write-host $texte
    $global:Form.Dispatcher.Invoke(
    [action]{$UI_textBlock_info.Text = $texte},
    "Render"
    )
}

# Form Code :
#===========================================================================

$UI_image_footer.Source = "$($PSCommandPathParent)PDW_footer.png"
$UI_textBox_login.Text = "$($env:UserDomain)\$($env:UserName)"

$UI_comboBox_auth.SelectedIndex = 0 # issue with xmlreader/writter for this attribute

# Combobox authentication type
$UI_comboBox_auth.Add_SelectionChanged({
    If ($UI_comboBox_auth.SelectedItem -eq "Windows authentication") {
        $UI_textBox_login.IsEnabled = $false
        $UI_passwordBox.IsEnabled = $false
    } else {
        $UI_textBox_login.IsEnabled = $true
        $UI_passwordBox.IsEnabled = $true
    }
    
    # MoveUp selected item :
    if ($UI_comboBox_auth.SelectedIndex -ne 0) {
        $Index = $UI_comboBox_auth.SelectedIndex+1
        $UI_comboBox_auth.Items.Insert(0,[string]$UI_comboBox_auth.SelectedItem)
        $UI_comboBox_auth.SelectedIndex = 0
        $UI_comboBox_auth.Items.RemoveAt($Index)
        $UI_comboBox_auth.SelectedIndex = 0
    }
})

# Destination Nav service change :
$UI_textBox_ServiceInstance.Add_TextChanged({
    # Extract server and instance name
    if ($UI_textBox_ServiceInstance.text.IndexOf("\") -eq -1) {
        $DestInstance_server = "localhost"
        $DestInstance_service = $UI_textBox_ServiceInstance.text
    } else {
        $DestInstance_server = $UI_textBox_ServiceInstance.text.split('\')[0]
        $DestInstance_service = $UI_textBox_ServiceInstance.text.split('\')[1]
    }
    if ($DestInstance_server -eq "localhost") {
        # Check if instance exists :
        $instance = $(Get-NAVServerInstance -ServerInstance $DestInstance_service)
        if (-Not $([string]::IsNullOrEmpty($instance))) {
            # Get local service configuration :
            $Server = Get-NAVServerConfiguration $DestInstance_service | ForEach-Object { $_.Value.split(';')[1] }
            $ServerDBInstance = Get-NAVServerConfiguration $DestInstance_service | ForEach-Object { $_.Value.split(';')[2] }
            if ($ServerDBInstance -ne "") {
                $Server = "$Server\$ServerDBInstance"
            }
            $UI_textBox_DestinationDatabase.Text = Get-NAVServerConfiguration $DestInstance_service | ForEach-Object { $_.Value.split(';')[3] }
            $UI_textBox_ServerName.Text = $Server
        } else {
            # blank field if instance not find :
            $UI_textBox_DestinationDatabase.Text = ""
            $UI_textBox_ServerName.Text = ""
        }
    } else {
        # Get remote service configuration
        # -- remote nav cmdlet not working
        <#If ($DestInstance_service -ne "") {
            If ($UI_comboBox_auth.SelectedItem.content -eq "Windows authentication") {
                Invoke-Command -Computer $DestInstance_server -ScriptBlock {Get-NAVServerConfiguration $args} -ArgumentList "$DestInstance_service" 
            } else {
                $SecurePass = ConvertTo-SecureString -AsPlainText $UserPass -Force
                $Credential = New-Object System.Management.Automation.PSCredential -ArgumentList $UserID,$SecurePass
                Invoke-Command -Computer $DestInstance_server -ScriptBlock {Get-NAVServerConfiguration $args} -credential $Credential -ArgumentList "$DestInstance_service" 
            }
        }#>
    } 
})

# Get backup file dialog button :
$UI_button_getbackupfile.Add_Click({
    [System.Reflection.Assembly]::LoadWithPartialName("System.windows.forms") | Out-Null
    
    $OpenFileDialog = New-Object System.Windows.Forms.OpenFileDialog
    $OpenFileDialog.initialDirectory = $OpenFileDialog.filename
    $OpenFileDialog.filter = "SQL Backup file (*.bak)| *.bak|All files (*.*)|*.*"
    $OpenFileDialog.ShowDialog() | Out-Null
    If ($OpenFileDialog.filename -ne "") {
        $UI_textBox_SourceFile.Text = $OpenFileDialog.filename
    }
})

# Show console button :
$UI_button_ShowConsole.Add_Click({
    if ($global:ConsoleShow) {
        $global:ConsoleShow = $false
        Hide-Console
    } else {
        $global:ConsoleShow = $true
        Show-Console
    } 
})



$UI_button_copy.Add_Click({
    [void][Reflection.Assembly]::LoadWithPartialName("Microsoft.SqlServer.SMO") 

    # ====== Getting db parameters ======
    $Server = $UI_textBox_ServerName.Text
    $DestDatabase = $UI_textBox_DestinationDatabase.Text
    $SourceLogin = $UI_textBox_login.Text
    $SourcePass = $UI_passwordBox.Text
    $SourceFilePath = $UI_textBox_SourceFile.Text
    $UserID = $UI_textBox_login.Text
    $UserPass = $UI_passwordBox.Text
    $BackupFileFullPath = $UI_textBox_SourceFile.Text

    # Extract server and instance name (for remote usage)
    <#if ($UI_textBox_ServiceInstance.text.IndexOf("\") -eq -1) {
        $DestInstance_server = "localhost"
        $DestInstance_service = $UI_textBox_ServiceInstance.text
    } else {
        $DestInstance_server = $UI_textBox_ServiceInstance.text.split('\')[0]
        $DestInstance_service = $UI_textBox_ServiceInstance.text.split('\')[1]
    }#>
    $DestInstance_service = $UI_textBox_ServiceInstance.text;

    #Check parameters :
    $error_occur = $false
    if ($SourceFilePath -eq "") {
        ShowMessageBox "You have to set the full backup file path to restore" 0
        $error_occur = $true
    }
    if ($DestDatabase -eq "") {
        ShowMessageBox "You have to set the destination database name" 0
        $error_occur = $true
    }
    if ($DestInstance_service -eq "") {
        ShowMessageBox "You have to set the destination Nav service instance to run the process" 0
        $error_occur = $true
    }

    if ($error_occur -eq $false) {
        write-host "STARTING NAV DB COPY" -ForegroundColor Cyan
        $dt = Get-Date -Format yyyyMMddHHmmss
        New-Item -ItemType directory -Path $PSCommandPathParent$dt
        $error_ImportSetting = $false
        $error.Clear()

        # ====== Start service instance ======
        RefreshStatus("Start NAV Test service instance if not start...")
        Set-NAVServerInstance -ServerInstance $DestInstance_service -Start

        # ====== Get First Company ========
        RefreshStatus("Get company...")
        $ListCompanies =  Get-NAVCompany -ServerInstance $DestInstance_service
        $FirstCompany = $ListCompanies[0]

        # ====== Export all Objects from test ========
        RefreshStatus("Export all Objects from test...")
        $exportfinsqlcommand = """$NavIde"" command=exportobjects,file=$PSCommandPathParent$dt\AllObj.fob,servername=.,database=$DestDatabase,Logfile=$PSCommandPathParent$dt\ExportFob.txt"
        Write-Host $exportfinsqlcommand -ForegroundColor DarkGray
        cmd /c $exportfinsqlcommand
        if (-not $(Test-Path "$PSCommandPathParent$dt\AllObj.fob")) {
            write-host "Unable to export all fob objects" -ForegroundColor Red
            write-host $_.Exception.Message -ForeGround red
            $UI_textBlock_info.Text = "Unable to export all objects from test. Check your Nav Service credential."    
            $error_occur = $true
        }

        if ($error_occur -eq $false) {
            # ====== Export Settings Data ========
            RefreshStatus("Export Settings data...")
            Invoke-NAVCodeunit -ServerInstance $DestInstance_service -CompanyName $FirstCompany -CodeunitId 11218034 -MethodName "FctExportSettings" | out-null
            #if ($error -ne "") {
            #    ShowMessageBox "The Import/Export settings CodeUnit 11218034 did not exists on this instance or wasn't run correctly. The copy will not proceed." 1
            #}
            
            # ====== Stop service instance ======
            RefreshStatus("Stop NAV service instance...")
            Set-NAVServerInstance -ServerInstance $DestInstance_service -Stop

            # ====== CONNECTION TO DATABASE ======
            $UI_textBlock_info.Text = "Connecting to dabatase..."
            write-host "Connecting to database..."
            $cn = New-Object System.Data.SQLClient.SQLConnection 
            # Windows authentification
            If ($UI_comboBox_auth.SelectedItem -eq "Windows authentication") {
                $cn.ConnectionString = "Server= $Server;Trusted_Connection=true;Connection Timeout=2;"
            # Local trusted connection
            } else {
                $cn.ConnectionString = "Server= $Server;user id=$UserID;password=$UserPass;Trusted_Connection=false;Connection Timeout=2;"
            }
            # Validate connection
            $cn.Open()
            write-host "Connection state = $($cn.State)"
            if (-not ($cn.State -eq "Open")) {
                ShowMessageBox "Unable to connect to database please check credential, server name, security policy and disable your vpn" 1
            } else {
                $cmd = new-object System.Data.SqlClient.SqlCommand 
                $cmd.Connection = $cn
                $cmd.CommandTimeout = 600
            
                # ====== GET DATABASE LOGICAL FILE NAME =====
                RefreshStatus("Get BACKUP FILE LOGIC NAME")
                $sql = "RESTORE FILELISTONLY FROM DISK = N'$BackupFileFullPath'"
                write-host $sql -ForegroundColor DarkGray
                $cmd.CommandText = $sql
                $Reader = $cmd.ExecuteReader()
                while ($Reader.Read()) {
                    if ($Reader['Type'] -eq "D") {
                        $LogicalDataName = $Reader['LogicalName']
                    } else {
                        $LogicalLogName = $Reader['LogicalName']
                    }
                }
                $Reader.Close()
                Write-Host "Logical Data file name : $LogicalDataName"
                Write-Host "Logical Log file name : $LogicalLogName"

                # ====== GET SQL DEFAULT DIRECTORY ======
                RefreshStatus("Get default SQL file folder...")
                $sql = "SELECT SERVERPROPERTY('InstanceDefaultDataPath')"
                write-host $sql -ForegroundColor DarkGray
                $cmd.CommandText = $sql
                $Reader = $cmd.ExecuteReader()
                while ($Reader.Read()) {
                    $BackupDirectory = $Reader.GetString(0)
                }
                $Reader.Close()
                Write-Host "Sql folder path : $BackupDirectory"  
            
                # ====== BRING DATABASE OFFLINE =====
                RefreshStatus("Bring database single user for restore...")
                $sql = "USE [master] ALTER DATABASE [$DestDatabase] SET SINGLE_USER WITH ROLLBACK IMMEDIATE"
                write-host $sql -ForegroundColor DarkGray          
                $cmd.CommandText = $sql
                try {
                    $cmd.ExecuteNonQuery()
                } catch {
                    $UI_textBlock_info.Text = "Unable to bring SQL database single user for restore, see console for more details."
                    write-host $_.Exception.Message -ForeGround red
                }
                # ====== IMPORT BACKUP ======
                RefreshStatus("Restore database from backup...")
                Start-Sleep -m 10
                $sql = "USE [master] RESTORE DATABASE [$DestDatabase] FROM  DISK = N'$BackupFileFullPath' WITH  REPLACE, FILE = 1,  MOVE N'$LogicalDataName' TO N'$($BackupDirectory)\$($DestDatabase)_$($dt)_data.mdf',  MOVE N'$LogicalLogName' TO N'$($BackupDirectory)\$($DestDatabase)_$($dt)_Log.ldf',  NOUNLOAD"
                $sql = $sql -replace "\\", "\" # path fixing
                Write-Host $sql -ForegroundColor DarkGray
                $cmd.CommandText = $sql
                try {
                    $cmd.ExecuteNonQuery()
                } catch {
                    $UI_textBlock_info.Text = "SQL Restore failed, see console for more details."
                    write-host $_.Exception.Message -ForeGround red
                    $error_occur = $true
                }
                # ====== BRING DATABASE ONLINE =====
                RefreshStatus("Bring back database multi-user...")
                $sql = "USE [master] ALTER DATABASE [$DestDatabase] SET MULTI_USER"
                write-host $sql -ForegroundColor DarkGray          
                $cmd.CommandText = $sql
                try {
                    $cmd.ExecuteNonQuery()
                } catch {
                    $UI_textBlock_info.Text = "Unable to bring SQL database multi-user back, see console for more details."
                    write-host $_.Exception.Message -ForeGround red
                }

                if ($error_occur -eq $false) {
                    # ====== DELETE THE BACKUP FILE CREATED =====
                    if ($UI_checkBox.IsChecked) {
                        RefreshStatus("Deleting backup file...")
                        If ($UI_textBox_ServerName.Text -eq "LOCALHOST") {
                            Remove-Item $BackupFileFullPath   
                        } else {
                            If ($UI_comboBox_auth.SelectedItem.content -eq "Windows authentication") {
                                Invoke-Command -Computer $($Server.split('\')[0]) -ScriptBlock {Remove-Item -Path $args -force } -ArgumentList "$BackupFileFullPath" 
                            } else {
                                $SecurePass = ConvertTo-SecureString -AsPlainText $UserPass -Force
                                $Credential = New-Object System.Management.Automation.PSCredential -ArgumentList $UserID,$SecurePass
                                Invoke-Command -Computer $($Server.split('\')[0]) -ScriptBlock {Remove-Item -Path $args -force } -credential $Credential -ArgumentList "$BackupFileFullPath" 
                            }
                        }
                    }

                    # ====== Restart service instance ======
                    RefreshStatus("Convert nav database...")
                    Invoke-NAVDatabaseConversion -DatabaseName $DestDatabase -DatabaseServer $Server
                    RefreshStatus("Restart Nav service...")
                    Set-NAVServerInstance -ServerInstance $DestInstance_service -Restart

                    # ====== Import all Objects in test ========
                    RefreshStatus("Import all Objects in test...")
                    $ImportObjectsFunction = """$NavIde"" command=importobjects, file=""$PSCommandPathParent+$dt\AllObj.fob"", logfile=""$PSCommandPathParent$dt\ImportFob.log"", servername=".", database=""$DestDatabase"", ImportAction=overwrite"
                    Write-Host $ImportObjectsFunction -ForegroundColor DarkGray
                    cmd /c $ImportObjectsFunction
                    if ($(Test-Path "$PSCommandPathParent$dt\ImportFob.log")) {
                        $error_occur = $true
                    }
                
                    # ====== Restart service instance ======
                    RefreshStatus("Restart Nav service...")
                    Set-NAVServerInstance -ServerInstance $DestInstance_service -Restart

                    # ====== Reimport Settings Data ========
                    RefreshStatus("Import Settings Data...")
                    Invoke-NAVCodeunit -ServerInstance $DestInstance_service -CompanyName $FirstCompany -CodeunitId 11218034 -MethodName "FctImportSetting" | out-null
                    if ($Error -notlike "") {
                        $error_ImportSetting = $true
                    }

                    if ($error -eq $true) {
                        Write-Host $(Get-Content "$PSCommandPathParent$dt\ImportFob.log")
                        ShowMessageBox "The test fob object import failed, please check there is no locked objects than manually import file : $PSCommandPathParent$dt\AllObj.fob (show console to see log details)" 0
                    }
                }
                # Drop sql connection
                $cn.Close()

                # Remove fob file even if restore failed (if not checked)
                if (-not $UI_checkBox_fob.IsChecked) {
                    Remove-Item $PSCommandPathParent$dt\AllObj.fob -Force
                }

                # Remove copy sub folder folder if no error and fob not keeped
                if (-not $($UI_checkBox_fob.IsChecked -or $error_occur)) {
                    Remove-Item $PSCommandPathParent$dt -Force -Recurse
                }

                # Finish notification
                if ($error_occur) {
                    ShowMessageBox "The copy was executed but an error occur during the process, please check the console for details." 1
                } else {
                    if ($error_ImportSetting -eq $true) {
                        Show-Console
                        ShowMessageBox "The copy was executed but the ""Import/Export settings"" code unit 11218034 could not be run correctly. You have to quickly check partners, wms and IC settings on the test environement. This can make issue on produciton base." 1
                    } else {
                        Show-Console
                        ShowMessageBox "The copy process was succesfully executed.  Test environement should be operational." 0
                    }
                }
                $UI_textBlock_info.Text = "No copy pending" 
            }
        }
    }        
})

# On close save the form xaml back
#===========================================================================
$form.Add_Closed({
    $UI_textBlock_info.Text = "No copy pending" 
    $xaml = [Windows.Markup.XamlWriter]::Save($form)
    $xaml | Set-Content "$($PSCommandPathParent)UI_CopyDatabase.xaml"
})
 
# Shows the form
#===========================================================================
$global:Form.ShowDialog() | out-null
